#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Author: Eduard Trott
# @Date:   2015-09-16 11:45:16
# @Email:  etrott@redhat.com
# @Last modified by:   etrott
# @Last Modified time: 2016-01-19 13:30:40


import os
import sys
import re
from string import ascii_uppercase

import gspread
import pandas as pd
import numpy as np

from .utils import get_credentials
from .gfiles import get_file_id, get_worksheet

# FIXME: clarify scopes
SCOPES = ('https://www.googleapis.com/auth/drive.metadata.readonly '
          'https://www.googleapis.com/auth/drive '
          'https://spreadsheets.google.com/feeds '
          'https://docs.google.com/feeds')


def download(gfile, wks_name=None, col_names=False, row_names=False,
            credentials=None, start_cell = 'A1'):
    """
        Download Google Spreadsheet and convert it to Pandas DataFrame

        :param gfile: path to Google Spreadsheet or gspread ID
        :param wks_name: worksheet name
        :param col_names: assing top row to column names for Pandas DataFrame
        :param row_names: assing left column to row names for Pandas DataFrame
        :param credentials: provide own credentials
        :param start_cell: specify where to start capturing of the DataFrame; default is A1
        :type gfile: str
        :type wks_name: str
        :type col_names: bool
        :type row_names: bool
        :type credentials: class 'oauth2client.client.OAuth2Credentials'
        :type start_cell: str
        :returns: Pandas DataFrame
        :rtype: class 'pandas.core.frame.DataFrame'

        :Example:

            >>> from df2gspread import gspread2df as g2d
            >>> df = g2d.download(gfile="1U-kSDyeD-...", col_names=True, row_names=True)
            >>> df
                   col1 col2
            field1    1    2
            field2    3    4
    """

    # access credentials
    credentials = get_credentials(credentials)
    # auth for gspread
    gc = gspread.authorize(credentials)

    try:
        # if gfile is file_id
        gc.open_by_key(gfile).__repr__()
        gfile_id = gfile
    except:
        # else look for file_id in drive
        gfile_id = get_file_id(credentials, gfile)

    if gfile_id is None:
        raise RuntimeError(
            "Trying to open non-existent or inaccessible spreadsheet")

    wks = get_worksheet(gc, gfile_id, wks_name)

    if wks is None:
        raise RuntimeError(
            "Trying to open non-existent or inaccessible worksheet")

    raw_data = wks.get_all_values()

    if not raw_data:
        raise ValueError(
            'Worksheet is empty or invalid.'
            )

    start_row_int, start_col_int = gspread.utils.a1_to_rowcol(start_cell)

    rows, cols  = np.shape(raw_data)
    if start_col_int > cols or (row_names and start_col_int + 1 > cols):
        raise RuntimeError(
            "Start col (%s) out of the table columns(%s)" % (start_col_int +
                                                            row_names, cols))

    if start_row_int > rows or (col_names and start_row_int + 1 > rows):
        raise RuntimeError(
            "Start row (%s) out of the table rows(%s)" % (start_row_int +
                                                            col_names, rows))

    raw_data = [row[start_col_int-1:] for row in raw_data[start_row_int-1:]]

    if row_names and col_names:
        row_names = [row[0] for row in raw_data[1:]]
        col_names = raw_data[0][1:]
        raw_data = [row[1:] for row in raw_data[1:]]
    elif row_names:
        row_names = [row[0] for row in raw_data]
        col_names = np.arange(len(raw_data[0]) - 1)
        raw_data = [row[1:] for row in raw_data]
    elif col_names:
        row_names = np.arange(len(raw_data) - 1)
        col_names = raw_data[0]
        raw_data = raw_data[1:]
    else:
        row_names = np.arange(len(raw_data))
        col_names = np.arange(len(raw_data[0]))

    df = pd.DataFrame([pd.Series(row) for row in raw_data], index=row_names)
    df.columns = col_names

    return df



import requests


class ForeignExchangeRatesConverter:
    def get_exchange_by_date(self, date, base_currency='EUR'):
        """
        :param date: date for currency conversion rates in %Y-%m-%d
        :param base_currency: currency in iso format 3 to discover conversions
        :return: json response of currency conversion
        """
        url = "https://api.exchangeratesapi.io/{date}?base={base_currency}" \
            .format(date=date, base_currency=base_currency)
        response = requests.request("GET", url)
        json_response = {}

        if response.ok:
            json_response = response.json().get('rates', {})
            for currency in json_response.keys():
                json_response[currency] = 1 / json_response[currency]

            json_response['base_currency'] = 'EUR'
            json_response['date'] = date

        return json_response



import tempfile

import google
from retry import retry

from gcloud_connectors.logger import EmptyLogger
from google.cloud import storage
import os
import json


class GStorageConnector:
    def __init__(self, confs_path=None, auth_type='service_account', json_keyfile_dict=None, logger=None):
        self.confs_path = confs_path
        self.json_keyfile_dict = json_keyfile_dict
        self.auth_type = auth_type

        # authorization boilerplate code

        # os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = r'{}/{}'.format(os.getcwd(), confs_path)

        if self.confs_path is None:

            from tempfile import NamedTemporaryFile
            with NamedTemporaryFile('w') as jsonfile:
                json.dump(json_keyfile_dict, jsonfile)
                jsonfile.flush()

                self.service = storage.Client.from_service_account_json(jsonfile.name)
        else:
            self.service = storage.Client.from_service_account_json(self.confs_path)
        self.logger = logger if logger is not None else EmptyLogger()

    def pd_to_gstorage(self, df, bucket_name, file_name_path):
        """
        :param df: pandas DataFrame to be saved on GCS
        :param bucket_name: GCS bucket name
        :param file_name_path: path to save file on bucket
        :return: True, False whether file is correctly saved or not
        """
        bucket = self.service.get_bucket(bucket_name)
        with tempfile.NamedTemporaryFile('w') as temp:
            df.to_parquet(temp.name + '.parquet', index=False)
            bucket.blob(file_name_path).upload_from_filename(temp.name + '.parquet', content_type='application/octet-stream')
            temp.flush()
            os.remove(temp.name + '.parquet')
            return True
        return False

    def recursive_delete(self, bucket_name, directory_path_to_delete):
        """
        :param bucket_name: GCS bucket name
        :param directory_path_to_delete: path to start recursive deletion
        :return: list of deleted files from GSC
        """
        bucket = self.service.get_bucket(bucket_name)
        blobs = bucket.list_blobs(prefix=directory_path_to_delete)
        deleted_files = []
        for blob in blobs:
            blob.delete()
            self.logger.info('deleted {}'.format(blob.name))
            deleted_files.append(blob.name)
        return deleted_files

    @retry((google.api_core.exceptions.GatewayTimeout), tries=3, delay=2)
    def copy_blob(self, source_bucket, dest_bucket, blob):
        return source_bucket.copy_blob(blob, dest_bucket, new_name=blob.name)

    def recursive_copy_between_buckets(self, source_bucket, dest_bucket, prefix, delimiter='/', to_delete=False):
        """

        :param source_bucket: source bucket where files are currently located
        :param dest_bucket: destination bucket where to copy files
        :param prefix: to filter based on path hierarchy
        :param delimiter: wildcard to match files
        :param to_delete: True if you want to delete blobs from source bucket, default is False
        :return:
        """
        source_bucket = self.service.get_bucket(source_bucket)
        dest_bucket = self.service.get_bucket(dest_bucket)

        blobs = source_bucket.list_blobs(prefix=prefix, delimiter=delimiter)  # assuming this is tested

        for blob in blobs:
            self.copy_blob(source_bucket=source_bucket, dest_bucket=dest_bucket, blob=blob)
            self.logger.info('copied {} from {} to {}'.format(blob.name, source_bucket.name, dest_bucket.name))
            if to_delete is True:
                source_bucket.delete_blob(blob.name)




import json

import googleapiclient
from googleapiclient.discovery import build
import httplib2
import pandas as pd
from datetime import datetime, timedelta
from oauth2client.service_account import ServiceAccountCredentials
from retry import retry

from gcloud_connectors.logger import EmptyLogger

SCOPES = ['https://www.googleapis.com/auth/analytics.readonly']


class GAError(Exception):
    """Base class for GA exceptions"""


class GAPermissionError(GAError):
    """Base class for GA permission exceptions"""
    pass


class GAnalyticsConnector:
    def __init__(self, confs_path=None, auth_type='service_accounts', json_keyfile_dict=None, logger=None):
        self.confs_path = confs_path
        self.json_keyfile_dict = json_keyfile_dict
        self.auth_type = auth_type

        if self.json_keyfile_dict is None:
            self.creds = ServiceAccountCredentials.from_json_keyfile_name(
                self.confs_path, scopes=SCOPES)
        else:
            self.creds = ServiceAccountCredentials.from_json_keyfile_dict(
                self.json_keyfile_dict, scopes=SCOPES)

        self.http = self.creds.authorize(httplib2.Http())

        self.service = build('analytics', 'v4', http=self.http,
                             discoveryServiceUrl='https://analyticsreporting.googleapis.com/$discovery/rest',
                             cache_discovery=False)
        self.logger = logger if logger is not None else EmptyLogger()
        self.management_service = None

    def get_segments_by_id(self, segment_id):
        self.management_service = build('analytics', 'v3', http=self.http)
        segments = self.management_service.management().segments().list().execute().get('items', [])
        for segment in reversed(segments):
            pass

    @retry(googleapiclient.errors.HttpError, tries=3, delay=2)
    def pd_get_report(self, view_id, start_date, end_date, metrics, dimensions, filters=None, page_size=100000,
                      page_token=None, comes_from_sampling=False, segments=None):
        """
        :param view_id: Google Analytics view id
        :param start_date: start_date to get data
        :param end_date: end_data to get data
        :param metrics: list of ga: metrics
        :param dimensions: list of ga: dimensions
        :param filters: filters in string type format. Ex. ga:deviceCategory==tablet
        :param page_size: max size of results per page
        :param page_token: token that identifies the page of results (depending on page_size)
        :param comes_from_sampling: to better debug unsampling capabilities, only with active logger initiated with GAnalyticsConnector
        :param segments: list of Google Analytics segments in format ['gaid::...', ...]
        :return: pandas DataFrame unsampled results from Google Analytics
        """
        self.logger.info('start view {view_id} from sampled data {comes_from_sampling} '
                         'from {start_date} to {end_date} '
                         'with metrics ({metrics}), dimensions ({dimensions}), filters ({filters}), page_token {page_token}'
                         .format(view_id=view_id, comes_from_sampling=comes_from_sampling,
                                 start_date=start_date, end_date=end_date, metrics=', '.join(metrics),
                                 dimensions=', '.join(dimensions),
                                 filters=filters, page_token=page_token))
        body = {
            'reportRequests': [
                {
                    'viewId': view_id,  # Add View ID from GA
                    'samplingLevel': 'LARGE',
                    'dateRanges': [{'startDate': start_date, 'endDate': end_date}],
                    'metrics': [{'expression': metric} for metric in metrics],
                    'dimensions': [{'name': dimension} for dimension in dimensions],  # Get Pages
                    # 'orderBys': [{"fieldName": "ga:sessions", "sortOrder": "DESCENDING"}],
                    'pageSize': page_size
                }]
        }
        if filters is not None:
            body['reportRequests'][0]['filtersExpression'] = filters
        if page_token is not None:
            body['reportRequests'][0]['pageToken'] = page_token
        if segments is not None:
            body['reportRequests'][0]["segments"] = [{"segmentId": segment} for segment in segments] if not isinstance(
                segments[0], dict) else segments
            body['reportRequests'][0]["dimensions"].append({"name": "ga:segment"})

        response = self.service.reports().batchGet(body=body).execute()

        df = self.get_df_from_response(response, dimensions, metrics)
        if response['reports'][0]['data'].get('samplesReadCounts') is not None:
            self.logger.info('unsampling for {} {}'.format(start_date, end_date))
            # difference of start_date minus end_date in days, then split api of api calls in two parts:
            # start_date until half difference in days
            # half difference in days plus one day until end_date
            start_date_dt = datetime.strptime(start_date, '%Y-%m-%d')
            end_date_dt = datetime.strptime(end_date, '%Y-%m-%d')
            middle_dt = start_date_dt + (end_date_dt - start_date_dt) / 2
            if end_date_dt > start_date_dt:
                if middle_dt == end_date_dt:
                    middle_dt = start_date_dt
                df = self.pd_get_report(view_id, start_date_dt.strftime('%Y-%m-%d'), middle_dt.strftime('%Y-%m-%d'),
                                        metrics, dimensions,
                                        filters=filters, page_size=page_size, page_token=page_token, segments=segments,
                                        comes_from_sampling=True)

                middle_dt_plus_1 = (middle_dt + timedelta(days=1))
                middle_dt_plus_1 = middle_dt_plus_1 if middle_dt_plus_1 <= end_date_dt else end_date_dt
                df = df.append(
                    self.pd_get_report(view_id, middle_dt_plus_1.strftime('%Y-%m-%d'), end_date_dt.strftime('%Y-%m-%d'),
                                       metrics, dimensions,
                                       filters=filters, page_size=page_size, page_token=page_token, segments=segments,
                                       comes_from_sampling=True))

        else:
            self.logger.info('no sample for {} {}'.format(start_date, end_date))
            next_page_token = response['reports'][0].get('nextPageToken')
            if next_page_token is not None:
                df = df.append(self.pd_get_report(view_id, start_date, end_date, metrics, dimensions,
                                                  filters=filters, page_size=page_size, page_token=next_page_token,
                                                  segments=segments))

        return df

    @retry(googleapiclient.errors.HttpError, tries=3, delay=2)
    def pd_get_raw_report(self, report_request, dimensions, metrics, page_size=10000, page_token=None,
                          comes_from_sampling=False):
        """
        Useful if combined with Export Reporting API v4 taken from Da Vinci Tools Chrome Extension https://chrome.google.com/webstore/detail/da-vinci-tools/pekljbkpgnpphbkgjbfgiiclemodfpen
        :param report_request: accepts only one report request at time to unsample correctly
        :param page_size: max size of results per page
        :param page_token: token that identifies the page of results (depending on page_size)
        :param comes_from_sampling: to better debug unsampling capabilities, only with active logger initiated with GAnalyticsConnector
        :return: pandas DataFrame unsampled results from Google Analytics
        :return:
        """
        start_date = report_request['reportRequests'][0]['dateRanges'][0]['startDate']
        end_date = report_request['reportRequests'][0]['dateRanges'][0]['endDate']
        report_request['reportRequests'][0]['pageSize'] = page_size
        if page_token is not None:
            report_request['reportRequests'][0]['pageToken'] = page_token
        response = self.service.reports().batchGet(body=report_request).execute()

        df = self.get_df_from_response(response, dimensions, metrics)

        if response['reports'][0]['data'].get('samplesReadCounts') is not None:
            self.logger.info('unsampling for {} {}'.format(start_date, end_date))
            # difference of start_date minus end_date in days, then split api of api calls in two parts:
            # start_date until half difference in days
            # half difference in days plus one day until end_date
            start_date_dt = datetime.strptime(start_date, '%Y-%m-%d')
            end_date_dt = datetime.strptime(end_date, '%Y-%m-%d')
            middle_dt = start_date_dt + (end_date_dt - start_date_dt) / 2
            if end_date_dt > start_date_dt:
                if middle_dt == end_date_dt:
                    middle_dt = start_date_dt
                report_request['reportRequests'][0]['dateRanges'][0]['startDate'] = start_date_dt.strftime('%Y-%m-%d')
                report_request['reportRequests'][0]['dateRanges'][0]['endDate'] = middle_dt.strftime('%Y-%m-%d')
                df = self.pd_get_raw_report(report_request,
                                            metrics, dimensions,
                                            comes_from_sampling=True)

                middle_dt_plus_1 = (middle_dt + timedelta(days=1))
                middle_dt_plus_1 = middle_dt_plus_1 if middle_dt_plus_1 <= end_date_dt else end_date_dt
                report_request['reportRequests'][0]['dateRanges'][0]['startDate'] = middle_dt_plus_1.strftime(
                    '%Y-%m-%d')
                report_request['reportRequests'][0]['dateRanges'][0]['endDate'] = end_date_dt.strftime('%Y-%m-%d')
                df = df.append(
                    self.pd_get_raw_report(report_request,
                                           metrics, dimensions,
                                           comes_from_sampling=True))
            else:
                self.logger.info('no sample for {} {}'.format(start_date, end_date))
                next_page_token = response['reports'][0].get('nextPageToken')
                if next_page_token is not None:
                    df = df.append(self.pd_get_raw_report(report_request, dimensions, metrics, page_size=10000,
                                                          page_token=None, comes_from_sampling=False))

            return df
        return df

    @staticmethod
    def get_df_from_response(response, dimensions, metrics):
        """
        :param response: raw response from Google Analytics API
        :param dimensions: list of Google Analytics dimensions
        :param metrics: list of Google Analytics metrics
        :return: pandas DataFrame of results from response
        """
        data_dic = {f"{i}": [] for i in dimensions + metrics}
        for report in response.get('reports', []):
            rows = report.get('data', {}).get('rows', [])
            for row in rows:
                for i, key in enumerate(dimensions):
                    data_dic[key].append(row.get('dimensions', [])[i])  # Get dimensions
                date_range_values = row.get('metrics', [])
                for values in date_range_values:
                    all_values = values.get('values', [])  # Get metric values
                    for i, key in enumerate(metrics):
                        data_dic[key].append(all_values[i])

        df = pd.DataFrame(data=data_dic)
        df.columns = [col.split(':')[-1] for col in df.columns]

        return df
